(* LazyLists with OCaml laziness *)
(* type t *)
type 'a t   = 'a cell Lazy.t
and 'a cell = Nil | Cons of ('a * 'a t)
(* fc : 'a Lazy.t -> 'a *)
let fc = Lazy.force
(* tl : 'a t -> 'a t *)
let tl xs = match fc xs with
 | Nil        -> failwith "empty list"
 | Cons(_,xs) -> xs
(* filter : ('a -> bool) -> 'a t -> 'a t *)
let rec filter p xs = lazy(match fc xs with
 | Nil        -> Nil
 | Cons(x,xs) -> if p x then Cons(x,filter p xs)
                        else fc(filter p xs)
)
(* to_list : int -> 'a t -> 'a list *)
let rec to_list n xs = if n < 1 then [] else match fc xs with
 | Nil        -> []
 | Cons(x,xs) -> x :: to_list (n-1) xs
(* zip_with : ('a -> 'b -> 'c) -> 'a t -> 'b t -> 'c t *)
let rec zip_with f xs ys = lazy(match (fc xs,fc ys) with
 | (Cons(x,xs),Cons(y,ys)) -> Cons(f x y,zip_with f xs ys)
 | _                       -> Nil
)
(* from : int -> int t *)
let rec from n = lazy(Cons(n,from(n+1)))
(* sieve : int t -> int t *)
let rec sieve xs = lazy(match fc xs with
 | Nil        -> Nil
 | Cons(x,xs) ->
  Cons(x,sieve(filter (fun y -> y mod x <> 0) xs))
)
(* primes : int t *)
let primes = sieve(from 2)
(* fibs : int t *)
let rec fibs = lazy(Cons(0,lazy(Cons(1,zip_with (+) fibs (tl fibs)))))
(* print_elts : (Format.formatter -> 'a -> unit) -> Format.formatter -> 'a t -> unit *)
let rec print_elts i pe fmt xs = if i < 1 then Format.printf "..." else match fc xs with
 | Nil        -> ()
 | Cons(x,xs) -> match fc xs with
  | Nil    -> pe fmt x
  | Cons _ -> Format.fprintf fmt "%a; %a" pe x (print_elts (i-1) pe) xs
(* toplevel_printer_int : Format.formatter -> int t -> unit *)
let toplevel_printer_int fmt =
 Format.fprintf fmt "[%a]%!" (print_elts 100 (fun fmt -> Format.fprintf fmt "%i"))

